#include<stdlib.h>
#include<string.h>


typedef unsigned long dword;
typedef unsigned int  uint;
typedef unsigned char byte;

#define LEVIATHAN_HEIGHT 16   // height of tree                        
#define TABLE_SIZE       256  // number of words in substitution table
#define NUM_KEYSETUP_PASSES 2 // number of passes in key setup  
#define LEVIATHAN_MIN_LEAF_NUM (1<<(LEVIATHAN_HEIGHT)) 

#define UNROLL_HEIGHT 2
#define LEVIATHAN_MAX_NODE_NUM (1<<(LEVIATHAN_HEIGHT-UNROLL_HEIGHT))
 
class Leviaphan{
	struct leviathan_key{
		dword F[256];                    // function from 8 bits to 32 bits 
		dword x_stack[LEVIATHAN_HEIGHT]; // stack for internal node state      
		dword y_stack[LEVIATHAN_HEIGHT];
		dword z_stack[LEVIATHAN_HEIGHT];
		int index;                      // index into stack                
		dword x,y,z;                   // intermediate node state         
		dword tree_num;                  // the value of x at the root      
	};
	leviathan_key L;
	void f(dword &x,dword &y,dword &z) {
		ff(x,y);
		z+=z;            
	}
	void ff(dword &x,dword &y) 
	{
		y=_rotr(y,8);
		x^=L.F[(byte)x];
		y^=L.F[(byte)y];
		y=_rotr(y,8);
		x=_rotl(x,8);
		x^=L.F[(byte)x];
		y^=L.F[(byte)y];
		x=_rotl(x,8);
	}
	
	void g(dword &x,dword &y,dword &z) {
		gg(x,y);
		z+=z+1;
	}  
	void gg(dword &x,dword &y) {
		x=~x;             
		x=_rotr(x,8);         
		x^=L.F[(byte)x];     
		y^=L.F[(byte)y];     
		x=_rotr(x,8);         
		y=_rotl(y,8);          
		x^=L.F[(byte)x];     
		y^=L.F[(byte)y];     
		y=_rotl(y,8);
	}  
	dword h(dword a,dword b) {return a^b;}
	void d(dword &x,dword &y,dword &z) {x+=z,y+=x,x+=y;}
public:
	dword Output(uint tree_num,uint leaf);
	dword Seek(uint tree_num,uint leaf_num);
	void Encrypt(dword *dest,int nodes);
	void EncryptOptimized(dword *dest,int nodes);
	void SetKey(const byte *key,uint bytes_in_key);
};



void Leviaphan::SetKey(const byte *key,uint bytes_in_key) 
{
	int i, j, k, index;
	dword tmp;
	for (i=0; i<TABLE_SIZE; i++) L.F[i] = 0;
    // Each iteration of this loop we form the permutation of one line
    // (and, incidentally, also permute previously formed lines)
    for (j=0; j<4; j++) {
		// Initialize the new line to the identity permutation, and
		// shift the existing lines over one
		for (i=0; i<TABLE_SIZE; i++) L.F[i]=L.F[i]*TABLE_SIZE+i;
		// Initialize index to a line-dependant value, so that the four
		// lines will get distinct permutations
		index=j;
		// Do key setup passes a number of times, using the new line as
		// the index.
		for (k=0; k<NUM_KEYSETUP_PASSES; k++) {
			for (i=0; i < TABLE_SIZE; i++) {
				index+=(key[i%bytes_in_key]+L.F[i]);
				index&=(TABLE_SIZE-1);
				tmp=L.F[i];
				L.F[i]=L.F[index];
				L.F[index]=tmp;
			}
		}
	}
	// Finally, set S0 equal to the xor of itself with the identity
	// permutation, so that (S0[x] ^ x) is a permutation.
	for (i=0; i<TABLE_SIZE; i++) L.F[i]^=i;
	// Walk down left side of tree to initialize on leaf zero.
	Seek(0,0);
}

dword Leviaphan::Seek(uint tree_num,uint leaf_num)
{
	int i;
	dword x,y,z;
	i=1<<(LEVIATHAN_HEIGHT-1); 
	x=tree_num;
	y=0;
	z=1;
	L.index=0;
	while(i>0) {
		d(x,y,z);
		if(i&leaf_num) g(x,y,z);  // right 
		else {
			// push data onto stack 
			L.x_stack[L.index]=x;
			L.y_stack[L.index]=y;
			L.z_stack[L.index]=z;
			L.index++;
			f(x,y,z);  // left 
		}
		i>>=1;
	}
	L.x=x; L.y=y; L.z=z;
	L.tree_num=tree_num;
	return h(x,y);
}


void Leviaphan::Encrypt(dword *dest,int nodes) 
{
	int index;
	dword x,y,z, *dest_max=dest+nodes;
	// set information for next node from stack 
	index=L.index;
	x=L.x; y=L.y; z=L.z;
	while(dest<dest_max) {
		// go left as far as possible, pushing data on the way 
		while(z<LEVIATHAN_MIN_LEAF_NUM) { 
			// apply diffusion function 
			d(x, y, z);
			// push data onto stack 
			L.x_stack[index]=x;
			L.y_stack[index]=y;
			L.z_stack[index]=z;
			index++;
			// compute left child 
			f(x,y,z);  
		}
		// output keystream, pop stack, and go right once 
		*dest++=h(x, y);
		index--;
		x=L.x_stack[index];
		y=L.y_stack[index];
		z=L.z_stack[index];
		// compute right child 
		g(x,y,z);  
	}
	// save current data in leviathan_key 
	L.index=index;
	L.x=x; L.y=y; L.z=z;
}

dword Leviaphan::Output(uint tree_num,uint leaf)
{
	int i; dword x,y,z;
	i=1<<(LEVIATHAN_HEIGHT-1); 
	x=tree_num;
	y=0;
	z=1;    
	while(i>0) {
		d(x,y,z);
		if(i&leaf) g(x,y,z);  // right 
		else f(x,y,z);  // left
		i>>=1;
	}
	return h(x, y);
}


void Leviaphan::EncryptOptimized(dword *dest,int nodes) 
{
	int i, index;
	dword x,y,z, *dest_max=dest+nodes;
	dword x1,y1;
#if (UNROLL_HEIGHT==2)
	dword x2,y2;
#endif
	// initialize state 
	i=1<<(LEVIATHAN_HEIGHT-1); 
	x=L.tree_num;
	y=0;
	z=1;
	index=0;
	while(i>UNROLL_HEIGHT) {
		d(x,y,z);
		// double z 
		z+=z;
		// push data onto stack
		L.x_stack[index]=x;
		L.y_stack[index]=y;
		L.z_stack[index]=z;
		index++;
		ff(x,y);  // left 
		i>>=1;
	}
	while(dest<dest_max) {
		// go left as far as possible, pushing data on the way 
		while(z<LEVIATHAN_MAX_NODE_NUM) { 
			// apply diffusion function 
			d(x, y, z);
			// double z 
			z += z;
			// push data onto stack 
			L.x_stack[index]=x;
			L.y_stack[index]=y;
			// compute left child 
			ff(x,y);  
			L.z_stack[index]=z;
			index++;
		}
		// compute remaining UNROLL_HEIGHT levels and output keystream 
#if   (UNROLL_HEIGHT==1)
		d(x, y, z);
		x1=x; y1=y; 
		ff(x,y);  
		*dest++^=h(x,y);
		gg(x1,y1);
		*dest++^=h(x1,y1);
#elif (UNROLL_HEIGHT==2)
		d(x,y,z);
		z+=z;
		x2=x; y2=y;     
		ff(x,y);
		d(x,y,z);
		x1=x; y1=y; 
		ff(x,y);  
		*dest++^=h(x,y);
		gg(x1, y1);
		*dest++^=h(x1,y1);
		gg(x2,y2);
		z++; d(x2,y2,z); z--;
		x1=x2; y1=y2; 
		ff(x2,y2);  
		*dest++^=h(x2,y2);
		gg(x1,y1);
		*dest++^=h(x1, y1);
#else 
#error "UNROLL_HEIGHT must be equal to 1 or 2!"
#endif
		// pop stack, and go right once 
		index--;
		x=L.x_stack[index];
		y=L.y_stack[index];
		// compute right child 
		gg(x,y);   
		z=L.z_stack[index]+1;
	}
	// save current data in leviathan_key 
	L.index=index;
	L.x=x; L.y=y; L.z=z;
}



void main()
{
	dword text[4];
	strcpy((char*)text,"Hello world!");
	Leviaphan obj;
	obj.SetKey((byte*)"qwerty",6);
	obj.Seek(0,0);
	obj.EncryptOptimized(text,4);
	obj.Seek(0,0);
	obj.EncryptOptimized(text,4);
}

